import {
  Controller,
  Get,
  Post,
  Body,
  Patch,
  Param,
  Delete,
  NotFoundException,
  HttpCode,
  BadRequestException,
} from '@nestjs/common';
import { ApiResponse } from '@nestjs/swagger';
import { EntityNotFoundError } from 'typeorm';
import { Paginate, PaginateQuery, Paginated } from 'nestjs-paginate';
import { ZgcService } from './zgc.service';
import { ZoneCreateDto } from './dto/zone-create.dto';
import { ZoneUpdateDto } from './dto/zone-update.dto';
import { Zone } from './entities/zone.entity';
import { GroupCreateDto } from './dto/group-create.dto';
import { GroupUpdateDto } from './dto/group-update.dto';
import { Group } from './entities/group.entity';
import { CategoryCreateDto } from './dto/category-create.dto';
import { CategoryUpdateDto } from './dto/category-update.dto';
import { Category } from './entities/category.entity';

@Controller('zgc')
export class ZgcController {
  constructor(private readonly service: ZgcService) {}

  private handleEntityNotFoundError(id: number, e: Error) {
    if (e instanceof EntityNotFoundError) {
      throw new NotFoundException(`Not found: id=${id}`);
    } else {
      throw e;
    }
  }

  // -------------------- Zone --------------------
  // --
  @Get('/zones')
  @ApiResponse({ type: [Zone] })
  findZones(@Paginate() query: PaginateQuery): Promise<Paginated<Zone>> {
    return this.service.findZones(query);
  }

  @Get('/zones/:id')
  @ApiResponse({ type: Zone })
  async findZone(@Param('id') id: number): Promise<Zone | null> {
    // NOTE: the + operator returns the numeric representation of the object.
    return this.service.findZone(+id).catch((err) => {
      this.handleEntityNotFoundError(id, err);
      return null;
    });
  }

  @Post('/zones')
  @ApiResponse({ type: [Zone] })
  createZone(@Body() zoneCreateDto: ZoneCreateDto) {
    return this.service.createZone(zoneCreateDto);
  }

  @Patch('/zones/:id')
  // Status 204 because no content will be returned
  @HttpCode(204)
  async updateZone(
    @Param('id') id: number,
    @Body() zoneUpdateDto: ZoneUpdateDto,
  ): Promise<void> {
    if (Object.keys(zoneUpdateDto).length === 0) {
      throw new BadRequestException('Request body is empty');
    }
    await this.service.updateZone(+id, zoneUpdateDto).catch((err) => {
      this.handleEntityNotFoundError(id, err);
    });
  }

  @Delete('/zones/:id')
  @HttpCode(204)
  async deleteZone(@Param('id') id: number): Promise<void> {
    await this.service.deleteZone(+id).catch((err) => {
      this.handleEntityNotFoundError(id, err);
    });
  }

  // -------------------- Group --------------------
  // --
  @Get('/groups')
  @ApiResponse({ type: [Group] })
  findGroups(@Paginate() query: PaginateQuery): Promise<Paginated<Group>> {
    return this.service.findGroups(query);
  }

  @Get('/groups/:id')
  @ApiResponse({ type: Group })
  async findGroup(@Param('id') id: number): Promise<Group | null> {
    // NOTE: the + operator returns the numeric representation of the object.
    return this.service.findGroup(+id).catch((err) => {
      this.handleEntityNotFoundError(id, err);
      return null;
    });
  }

  @Post('/groups')
  @ApiResponse({ type: [Group] })
  createGroup(@Body() zoneCreateDto: GroupCreateDto) {
    return this.service.createGroup(zoneCreateDto);
  }

  @Patch('/groups/:id')
  // Status 204 because no content will be returned
  @HttpCode(204)
  async updateGroup(
    @Param('id') id: number,
    @Body() zoneUpdateDto: GroupUpdateDto,
  ): Promise<void> {
    if (Object.keys(zoneUpdateDto).length === 0) {
      throw new BadRequestException('Request body is empty');
    }
    await this.service.updateGroup(+id, zoneUpdateDto).catch((err) => {
      this.handleEntityNotFoundError(id, err);
    });
  }

  @Delete('/groups/:id')
  @HttpCode(204)
  async deleteGroup(@Param('id') id: number): Promise<void> {
    await this.service.deleteGroup(+id).catch((err) => {
      this.handleEntityNotFoundError(id, err);
    });
  }

  // -------------------- Category --------------------
  // --
  @Get('/categories')
  @ApiResponse({ type: [Category] })
  findCategories(
    @Paginate() query: PaginateQuery,
  ): Promise<Paginated<Category>> {
    return this.service.findCategories(query);
  }

  @Get('/categories/:id')
  @ApiResponse({ type: Category })
  async findCategory(@Param('id') id: number): Promise<Category | null> {
    // NOTE: the + operator returns the numeric representation of the object.
    return this.service.findCategory(+id).catch((err) => {
      this.handleEntityNotFoundError(id, err);
      return null;
    });
  }

  @Post('/categories')
  @ApiResponse({ type: [Category] })
  createCategory(@Body() categoryCreateDto: CategoryCreateDto) {
    return this.service.createCategory(categoryCreateDto);
  }

  @Patch('/categories/:id')
  // Status 204 because no content will be returned
  @HttpCode(204)
  async updateCategory(
    @Param('id') id: number,
    @Body() categoryUpdateDto: CategoryUpdateDto,
  ): Promise<void> {
    if (Object.keys(categoryUpdateDto).length === 0) {
      throw new BadRequestException('Request body is empty');
    }
    await this.service.updateCategory(+id, categoryUpdateDto).catch((err) => {
      this.handleEntityNotFoundError(id, err);
    });
  }

  @Delete('/categories/:id')
  @HttpCode(204)
  async deleteCategory(@Param('id') id: number): Promise<void> {
    await this.service.deleteCategory(+id).catch((err) => {
      this.handleEntityNotFoundError(id, err);
    });
  }
}
